def annual_query(df):
  return df["subscription.plan.interval"] == 'year'

def time_range_query(df,start_time,end_time):
    return ((df.created >= start_time) & (df.created < end_time))  & (df.amount_due > 0)

def billing_status_query(df):
    return (df.status != 'draft') & (df.status != 'void')

# query filters
def by_day_query(df,start_time,end_time,manual=False):
    return time_range_query(df,start_time,end_time) & (df.billing_reason != "manual" if not manual else df.billing_reason == "manual") & billing_status_query(df)



import pandas as pd
import requests
import datetime
import os
from datetime import datetime
from datetime import date

import petaldata
from petaldata import util
from petaldata.storage import *

class Resource(object):
  def __init__(self,base_pickle_filename="stripe_invoice.pkl"):
    self.csv_filename = None
    self.df = None
    self.__metadata = None
    self.local = Local(base_pickle_filename)
    self.s3 = S3.if_enabled(base_pickle_filename)

  def load(self):
    print("Attempting to load saved pickle file...")
    if (self.df is None) & (self.local.enabled == True): self.df = self.local.read_pickle_dataframe() 
    if (self.df is None) & (self.s3 is not None): self.df = self.s3.read_pickle_dataframe() 

    if self.df is None:
      print("\tNo cached files exist.")
      self.download()
      self.load_from_download()

    if self.df is None:
      print("\tUnable to load dataframe.")
    else:
      print("\t...Done. Dataframe Shape:",self.df.shape)
      return self.df

  # TODO - rename to merge?
  # https://en.wikipedia.org/wiki/Merge_(SQL)
  def update(self, created_gt=None):
    print("Updating...")
    if created_gt is None:
      print("\tSetting created_gt=",self.updated_at)
      created_gt = self.updated_at
    else:
      print("\tcreated >",created_gt)
    new = type(self)()
    new.download(created_gt=created_gt)
    new.load_from_download() # don't want to save a pickle! would override existing.
    self.upsert(new)
    return self

  def upsert(self,other_resource):
    old_count = self.df.shape[0]
    # only concat if rows found, otherwise dtypes can change
    # https://stackoverflow.com/questions/33001585/pandas-dataframe-concat-update-upsert
    if other_resource.df.shape[0] > 0:
      print("\tInserting new rows")
      print("created before concat:",self.df.created.dtype,self.df.created.head(5))
      print("other created before concat:",other_resource.df.created.dtype,other_resource.df.created.head(5))
      df = pd.concat([self.df, other_resource.df[~other_resource.df.index.isin(self.df.index)]])
      print("created after concat:",df.created.dtype,df.created.head(5))
      print("\tUpdating existing rows")
      df.update(other_resource.df)
      # timezone is stripped after update. add back.
      df = self.set_date_tz(df)
      print("created after set_date_z:",df.created.dtype,df.created.head(5))
      self.df = df

    new_count = self.df.shape[0] - old_count
    if new_count > 0:
      print("Added {} new rows. Now with {} total rows.".format(new_count, self.df.shape[0]))
      print("\tTime Range:",self.df[self.CREATED_AT_FIELD].min(), "-", self.df[self.CREATED_AT_FIELD].max())
      # TODO - I think hubspot saves after update ... make consistent
    else:
      print("No new rows.")
    return self
    

  @property
  def updated_at(self):
    return self.df[self.CREATED_AT_FIELD].max()

  def load_from_download(self,filename=None):
    if filename is None:
      filename = self.csv_filename
    else:
      filename = self.local.dir + filename
    print("Loading {} MB CSV file...".format(Local.file_size_in_mb(filename)))
    dataframe = pd.read_csv(filename,parse_dates = self.metadata.get("convert_dates"))
    dataframe.set_index(self.metadata.get("index"),inplace=True)
    print("created from download:",dataframe.created.dtype,dataframe.created.head(3))
    self.df = dataframe
    print("\t...Done. Dataframe Shape:",self.df.shape)
    count = self.df.shape[0]
    if count > 0:
      print("\tTime Range:",self.df[self.CREATED_AT_FIELD].min(), "-", self.df[self.CREATED_AT_FIELD].max())
    return self.df

  def request_params(self,created_gt,_offset):
    pass

  def reset(self):
    print("Resetting...")
    if self.local.enabled == True: self.local.delete()
    if self.s3: self.s3.delete()

    self.df = None
    self.__metadata = None
    return self

  def save(self):
    print("Saving to Pickle file...")
    if self.local.enabled == True: self.local.save(self)
    if self.s3: self.s3.save(self)

    return True

  def download(self,created_gt=None,_offset=None):
    first_chunk = True
    start_time = datetime.now()
    filename = self.local.csv_download_filename(self.CSV_FILE_PREFIX,start_time)
    print("Starting download to",filename,"...")
    with requests.get(self.api_url+".csv", headers=self.request_headers, params=self.request_params(created_gt,_offset), stream=True) as r:
        r.raise_for_status()
        with open(filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=None):
                # TODO - how to handle 500 errors in chunks?
                if chunk: # filter out keep-alive new chunks
                    if (first_chunk):
                        first_chunk = False
                        print("\t...will update progress every 25 MB of data transfer.")
                    f.write(chunk)
                    size_in_mb = Local.file_size_in_mb(filename)
                    if ((size_in_mb > 1) & (len(chunk) > 1000) & (size_in_mb % 25 == 0) ):
                        print("\tDownloaded", size_in_mb, "MB...")

    size_in_mb = Local.file_size_in_mb(filename)
    time_delta = datetime.now() - start_time
    print("\t...Done.\n\tFile Size=",size_in_mb,"MB." " Total Time=", 
          round(time_delta.seconds/60.0,2), "minutes", "\n\tLocation:", filename)
    self.csv_filename = filename
    return self.csv_filename

  @property
  def metadata(self):
    if self.__metadata is None:
      self.__metadata = self.get_metadata()

    return self.__metadata

  def get_metadata(self):
    r = requests.get(self.api_url+"/metadata.pandas",  headers=self.request_headers)
    r.raise_for_status()
    metadata = r.json()
    print("Loaded metadata w/keys=",list(metadata.keys()))
    return metadata

  @staticmethod
  def set_date_tz(dataframe):
    for col in dataframe.columns:
      # this is a tz-naive timestamp. w/a tz would look like "datetime64[ns, UTC]"
      if dataframe[col].dtype == 'datetime64[ns]':
        dataframe[col] = dataframe[col].dt.tz_localize('UTC')
        # Strips out tz info so don't have to worry about comparing tz-aware datetimes
        # dataframe[col] = dataframe[col].dt.tz_convert(None)
    return dataframe

  @property
  def api_url(self):
    return petaldata.api_base + self.RESOURCE_URL


import pandas as pd
import requests
import calendar
import datetime
import os
from datetime import datetime
from datetime import date

import petaldata
from petaldata import util

class Contact(object):
  RESOURCE_URL = '/hubspot/contacts'

  def __init__(self,pickle_filename="hubspot_contact.pkl"):
    self.csv_filename = None
    self.pickle_filename = petaldata.cache_dir + pickle_filename
    self.df = None
    self.__metadata = None
    pass

  @property
  def api_url(self):
    return petaldata.api_base + self.RESOURCE_URL

  @classmethod
  def from_pickle(cls, filename):
    cache_dir = os.path.dirname(os.path.realpath(filename))
    petaldata.cache_dir = cache_dir
    instance = cls(pickle_filename=filename)
    instance.load_dataframe_from_pickle()

    return instance

  def pickle_file_exists(self):
    return os.path.isfile(self.pickle_filename)

  @property
  def metadata(self):
    if self.__metadata is None:
      self.__metadata = self.get_metadata()

    return self.__metadata

  def get_metadata(self):
    r = requests.get(self.api_url+"/metadata.pandas",  headers=self.request_headers)
    r.raise_for_status()
    metadata = r.json()
    print("Loaded Hubspot metadata w/keys=",list(metadata.keys()))
    return metadata

  @property
  def request_headers(self):
    return {
      "Authorization": "Bearer %s" % (petaldata.api_key,),
      'HUBSPOT_API_KEY': petaldata.resource.hubspot.api_key
    }

  def load(self):
    # TODO - attempt to load from CSV file, which could happen with an incomplete download
    self.load_dataframe_from_pickle()
    if self.df is None:
      self.download()
      self.load_from_download()

  def update(self):
    contact_new = Contact()
    contact_new.download(created_gt=self.updated_at)
    contact_new.load_from_download() # don't want to save a pickle! would override existing.
    self.df = pd.concat([self.df,contact_new.df], verify_integrity=True)
    new_contacts_count = contact_new.df.shape[0]
    if new_contacts_count > 0:
      print("Added {} new contacts. Now with {} total contacts.".format(new_contacts_count, self.df.shape[0]))
      print("Time Range:",self.df.createdate.min(), "-", self.df.createdate.max())
      self.save()
    else:
      print("No new contacts.")
    return self

  def save(self):
    self.df.to_pickle(self.pickle_filename)
    print("Saved Dataframe to pickle. Size (MB)=",util.file_size_in_mb(self.pickle_filename))
    return self.pickle_filename

  def load_dataframe_from_pickle(self):
    if not self.pickle_file_exists():
      print("Pickle file does not exist at",self.pickle_filename)
      return None
    print("Loading {} MB Pickle file...".format(util.file_size_in_mb(self.pickle_filename)))
    print("\t...",self.pickle_filename)
    self.df = pd.read_pickle(self.pickle_filename)
    print("\t...Done. Dataframe Shape:",self.df.shape)
    print("\tTime Range:",self.df.createdate.min(), "-", self.df.createdate.max())
    return self.df

  @property
  def updated_at(self):
    return self.df.createdate.max()

  def load_from_download(self):
    print("Loading {} MB CSV file...".format(util.file_size_in_mb(self.csv_filename)))
    dataframe = pd.read_csv(self.csv_filename,parse_dates = self.metadata.get("convert_dates"))
    dataframe.set_index(self.metadata.get("index"),inplace=True)
    dataframe = self.__set_date_tz(dataframe)
    self.df = dataframe
    print("\t...Done. Dataframe Shape:",self.df.shape)
    contacts_count = self.df.shape[0]
    if contacts_count > 0:
      print("\tTime Range:",self.df.createdate.min(), "-", self.df.createdate.max())
    return self.df

  def __csv_download_filename(self,start_time):
    return petaldata.cache_dir + "hubspot_contact_" + start_time.strftime("%Y%m%d-%H%M%S") + ".csv"

  def download(self,created_gt=None,_offset=None):
    first_chunk = True
    start_time = datetime.now()
    filename = self.__csv_download_filename(start_time)
    with requests.get(self.api_url+".csv", headers=self.request_headers, params=self.__request_params(created_gt,_offset), stream=True) as r:
        r.raise_for_status()
        with open(filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=None):
                # TODO - how to handle 500 errors in chunks?
                if chunk: # filter out keep-alive new chunks
                    if (first_chunk):
                        first_chunk = False
                        print("Starting download...will update progress every 25 MB of data transfer.")
                        print("\tSaving to", filename)
                    f.write(chunk)
                    size_in_mb = util.file_size_in_mb(filename)
                    if ((size_in_mb > 1) & (len(chunk) > 1000) & (size_in_mb % 25 == 0) ):
                        print("\tDownloaded", size_in_mb, "MB...")

    size_in_mb = util.file_size_in_mb(filename)
    time_delta = datetime.now() - start_time
    print("\t...Done.\n\tFile Size=",size_in_mb,"MB." " Total Time=", 
          round(time_delta.seconds/60.0,2), "minutes", "\n\tLocation:", filename)
    self.csv_filename = filename
    return self.csv_filename

  def __set_date_tz(self,dataframe):
    for col in self.metadata.get("convert_dates"):
        if dataframe[col].count() > 0:
            # Hubspot sends down tz already (ex: "2019-04-22T00:00:00Z")
            # Strips out tz info so don't have to worry about comparing tz-aware datetimes
            # df[col] = pd.to_datetime(df_payments[col], utc=True).dt.tz_convert(None)
            dataframe[col] = dataframe[col].dt.tz_convert(None)
    return dataframe

  def __request_params(self,created_gt,_offset):
    params = {}
    if created_gt is not None:
       params['created_gt'] = calendar.timegm(created_gt.timetuple())
    if _offset is not None:
        params['_offset'] = _offset
    return params



import pandas as pd
import datetime
from datetime import datetime
from datetime import date

import petaldata
from petaldata.resource.stripe.reports import query_filters

class AbstractStripeReport(object):
  def __init__(self,invoices):
    self.df = invoices.df
    self.convert_annual_subs_to_monthly(self.df)
    self.df = self.add_simulated_annual_invoices(self.df)

  # def to_frame(self,tz='UTC'):
  #   self.df = self.set_frame_tz(self.df,tz=tz)

  @staticmethod
  def set_frame_tz(df,tz = 'UTC'):
    for col in df.columns:
      if 'datetime64[ns' in str(df[col].dtype):
        df[col] = df[col].dt.tz_convert(tz)
    return df

  @staticmethod
  def strip_frame_tz(df):
    print("Stripping timezone from datetime columns.")
    df = df.copy()
    for col in df.columns:
          if 'datetime64' in str(df[col].dtype):
            df[col] = df[col].apply(lambda x:datetime.replace(x,tzinfo=None))
    return df

  @staticmethod
  def setup_time(dt,tz=None):
    t=pd.Timestamp(dt)
    t=t.tz_localize(tz).tz_convert(None)
    return t

  @staticmethod
  def convert_annual_subs_to_monthly(df):
    print("Converting annual subscriptions to monthly")
    per_month = lambda amount, interval : amount / 12.0 if (interval == 'year') else amount
    df['amount_due_per_month'] = df.apply(lambda row: per_month(row.amount_due,row["subscription.plan.interval"]), axis=1)
    df['amount_paid_per_month'] = df.apply(lambda row: per_month(row.amount_paid,row["subscription.plan.interval"]), axis=1)

  @staticmethod
  def add_simulated_annual_invoices(df):
    print("Adding simulated monthly invoices from annual invoices...")
    frames = []
    for index, row in df[query_filters.annual_query(df) & (df.amount_due > 0)].iterrows():
        for datetime in pd.date_range(start=row["subscription.current_period_start"], 
                                  end=row["subscription.current_period_end"]-pd.offsets.MonthBegin(2), 
                                  freq='MS'):
            date = datetime.date()
            simulated_row = row.copy()
            simulated_row.name = row.name+"_"+str(date)
            t=pd.Timestamp(datetime)
            simulated_row.date = t
            simulated_row.created = t
            simulated_row["simulated"] = 1
            frame = simulated_row.to_frame().transpose()
            frame.index.name = "id"
            frames.append(frame)

    print("\t...concating")
    df_combined = pd.concat([df]+frames,verify_integrity=True,sort=False)
    print("\t...assigning dtypes")
    return AbstractStripeReport.assign_dtypes(df,df_combined)

  @staticmethod
  def assign_dtypes(original_df,other_df):
    return other_df.astype(original_df.dtypes.to_dict())


  @staticmethod
  def cents_to_dollars(df,cols=None):
    print("Converting cents to dollars. Cols=",cols)
    df[cols]=df[cols]/100
    return df

  def to_gsheet(df,spreadsheet_name=None,worksheet_name=None,service_file=None):
    gc=pygsheets.authorize(service_file=service_file)

    print("Opening Google Sheet...")

    # Must share sheet with "client_email" from JSON creds file
    sh = gc.open('Copy of Revenue Reporting')

    wks = sh.worksheet_by_title("Monthly MRR via Invoices")
    print("\t...updating MRR worksheet")
    wks.clear(fields="*")



    wks.set_dataframe(strip_tz(grouped),(1,1), copy_index=True, nan="")
    print("\t...Done.")





import os
from petaldata.storage.abstract_storage import AbstractStorage
from petaldata import util
import pandas as pd

class Local(AbstractStorage):
  enabled = True
  dir = None

  def available(self):
    return self.dir_exists()

  def dir_exists(self):
    return os.path.isdir(self.dir)

  def read_pickle_dataframe(self):
    if not self.pickle_file_exists():
      print("\tLocal Pickle file does not exist at",self.pickle_filename)
      return None
    print("\tLoading {} MB Pickle file...".format(self.file_size_in_mb(self.pickle_filename)))
    print("\t...",self.pickle_filename)
    self.df = pd.read_pickle(self.pickle_filename)
    return self.df

  def save(self,resource):
    print("\tSaving local file...")
    resource.df.to_pickle(self.pickle_filename)
    print("\t...Done. Size (MB)=",self.file_size_in_mb(self.pickle_filename))
    return self.pickle_filename

  def delete(self):
    if self.pickle_file_exists():
      print("\tDeleting local pickle file.\n\tSize (MB)=",self.file_size_in_mb(self.pickle_filename),"\n\t",self.pickle_filename)
      os.remove(self.pickle_filename)
    else:
      print("\tNo pickle file exists at",self.pickle_filename)

  @property
  def pickle_filename(self):
    return self.dir + self.base_pickle_filename

  def pickle_file_exists(self):
    return os.path.isfile(self.pickle_filename)

  @staticmethod
  def file_size_in_mb(filename):
    return AbstractStorage.bytes_to_mb(os.path.getsize(filename))

  def csv_download_filename(self,csv_file_prefix,start_time):
    return self.dir + csv_file_prefix + start_time.strftime("%Y%m%d-%H%M%S") + ".csv"



import petaldata
from petaldata.resource.abstract import Resource
import calendar

class Invoice(Resource):
  RESOURCE_URL = '/stripe/invoices'
  CREATED_AT_FIELD = 'created'
  CSV_FILE_PREFIX = "stripe_invoices_"

  @property
  def request_headers(self):
    return {
      "Authorization": "Bearer %s" % (petaldata.api_key,),
      'STRIPE-API-KEY': petaldata.resource.stripe.api_key
    }

  def request_params(self,created_gt,_offset):
    params = {}
    if created_gt is not None:
       params['created_gt'] = calendar.timegm(created_gt.timetuple())
    if _offset is not None:
        params['_offset'] = _offset
    return params



import boto3
import io
import smart_open
import pickle
import pandas as pd
from petaldata.storage.abstract_storage import AbstractStorage

class S3(AbstractStorage):
  enabled = False
  aws_access_key_id = None
  aws_secret_access_key = None
  bucket_name = None

  def __init__(self,base_pickle_filename):
    self.__client = None
    self.__resource = None
    super().__init__(base_pickle_filename)

  def available(self):
    return self.list_permissions();

  def read_pickle_dataframe(self):
    if not self.pickle_file_exists():
      print("\tS3 Pickle file does not exist at",self.pickle_uri())
      return None
    else:
      print("\tS3 Pickle exists at",self.pickle_uri())
      print("\tS3 Loading {} MB Pickle file...".format(self.file_size_in_mb(self.base_pickle_filename)))
      return self.stream_and_read_pickle()

  def save(self,resource):
    print("\tSaving S3 file...")
    pickle_buffer = io.BytesIO()
    resource.df.to_pickle(pickle_buffer,compression=None)
    obj = self.resource().Object(self.bucket_name, self.base_pickle_filename).put(Body=pickle_buffer.getvalue())
    print("\t...Done. Size (MB)=",self.file_size_in_mb(self.base_pickle_filename))
    return obj

  def lookup(self,key):
    return self.bucket().Object(key)

  def delete(self):
    if self.pickle_file_exists():
      obj = self.lookup(self.base_pickle_filename)
      print("S3 Deleting pickle file.\n\tSize (MB)=",self.bytes_to_mb(obj.content_length))
      return obj.delete()

  def file_size_in_mb(self,key):
    return self.bytes_to_mb(self.lookup(key).content_length)

  def session(self):
    return boto3.Session(
      aws_access_key_id=self.aws_access_key_id,
      aws_secret_access_key=self.aws_secret_access_key
    )

  def pickle_uri(self):
    return "s3://{}/{}".format(self.bucket_name,self.base_pickle_filename)

  def stream_and_read_pickle(self):
    f=smart_open.open(self.pickle_uri(), 'rb', transport_params=dict(session=self.session()))
    pickle.loads(f.read())
    return pd.read_pickle(f, compression=None)

  # https://stackoverflow.com/questions/47056628/verifying-s3-credentials-w-o-get-or-put-using-boto3
  def list_permissions(self):
    try:
      response = self.client().head_bucket(Bucket=self.bucket_name)
      # see response['ResponseMetadata']['HTTPStatusCode']
      return True
    except Exception as e:
      print("Error trying to fetch bucket=",self.bucket_name,"Error:",e)
      return False

  def client(self):
    if self.__client is None:
      self.__client = boto3.client('s3', aws_access_key_id=self.aws_access_key_id, aws_secret_access_key=self.aws_secret_access_key)
    return self.__client

  def resource(self):
    if self.__resource is None:
      self.__resource = boto3.resource('s3', aws_access_key_id=self.aws_access_key_id, aws_secret_access_key=self.aws_secret_access_key)
    return self.__resource

  def bucket(self):
    return self.resource().Bucket(self.bucket_name)

  def pickle_file_exists(self):
    return self.file_exists(self.bucket(),self.base_pickle_filename)

  @staticmethod
  def file_exists(bucket,key):
    objs = list(bucket.objects.filter(Prefix=key))
    if len(objs) > 0 and objs[0].key == key:
        return True
    else:
        return False


import pandas as pd
import datetime
from datetime import datetime
from datetime import date

import petaldata
from petaldata.resource.stripe.reports.abstract_stripe_report import AbstractStripeReport
from petaldata.resource.stripe.reports import query_filters

class MRRByMonth(AbstractStripeReport):
  def __init__(self,invoices):
    super().__init__(invoices)

  def to_frame(self,tz='UTC',end_time=datetime.now()):
    df = self.set_frame_tz(self.df,tz=tz)
    df = self.strip_frame_tz(self.df)
    end_timestamp = self.setup_time(end_time,tz=tz)

    df_filtered = df[(df.billing_reason != "manual") & query_filters.time_range_query(df,df.created.min(),end_timestamp) & query_filters.billing_status_query(df)]
    grouped = df_filtered.groupby(pd.Grouper(key="created",freq="M")).agg(
      {
        "amount_due": "sum", 
        "amount_paid": "sum",
        "amount_due_per_month": "sum", 
        "amount_paid_per_month": "sum",
        "simulated": "sum",
        "paid": 'sum',
        "created": 'count',
        "customer_email": pd.Series.nunique
      }
    )
    grouped.rename(columns={"customer_email": 'customers', "simulated": 'ongoing_annual_subscriptions'},inplace=True)
    grouped.sort_index(ascending=False)
    grouped = self.cents_to_dollars(grouped, cols=[i for i in grouped.columns if 'amount' in i])
    return grouped